{
         
        Width, Height: Integer
        CellLayer: array[0..Height-1] of array[0..Height-1] of Integer;
        ObjectLayer: array[0..Width-1] of array[0..Height-1] of Integer;
}
unit DGame;

interface

uses
  //GL,
  windows,

  DUtils,
  DDebug,
  DMath,
  DEvents,
  DWinApi,
  DSound,
  DGraph,
  DEditor;

const
  GWidth = 640;
  GHeight = 480;
  GCell = 32; //  

type
TAction = (
                noAction,
                dirRight,
                dirLeft,
                dirUp,
                dirDown,

                bomLeft,
                bomRight
          );

TGame = class;
TGameObject = class;
TPlayer = class;
TGameObjectClass = class of TGameObject;

TGame = class(TWinApp)
private
  FWinApi: TWinApi;
  FWindow: TWindow;
  FSound: TSound;
  FGraph: TGraph;

  FEvents: array Of TEvent;

  FLevel: string;

  FTex: TTexture;
  FSprite: TSprite;
  FWidth, FHeight: Integer;

  FPlayer: TPlayer;
  FCells: array of array of TGameObject;
  FObjects: array of TGameObject;

  FAi: Boolean;
  FDocs: Integer;
  
  FEditor: TEditor;
  procedure LoadLevel(LevelName: string);
public
  constructor Create;
  destructor Destroy; override;
  procedure Run;
  procedure RunEditor;
  //   
  function GetTimer: Integer; inline;
  procedure Log(Mess: string); inline;
  procedure Assert(Cond: Boolean; Mess: string = 'TGame.Assert error'); inline;
  procedure PlaySound(FileName: string); inline;
  procedure PlayMusic(FileName: string); inline;

  function GetWidth: Integer; inline;
  function GetHeight: Integer; inline;

  procedure BeginRender; inline;
  procedure Clear(R, G, B: Single; A: Single = 1); inline;
  procedure EndRender; inline;
  procedure SetBlendMode(Mode: TBlendMode); inline;
  function LoadTex(FileName: string): TTexture; inline;
  function CreateSprite: TSprite; inline;
  procedure SetTexture(ID: TTexture); inline;
  procedure PushMatrix; inline;
  procedure PopMatrix; inline;
  procedure Translate(V: TVec2f); inline;
  //
  procedure Render; override;
  procedure Update; override;
  procedure OnEvent(const Event: TEvent); override;
  // game
  procedure DoAi;
  procedure Lose;
  procedure Bomb(Cell: TVec2i);
  procedure DecDoc;
  function IsValidCell(Pos: TVec2i): Boolean;
  function FreeCell(Pos: TVec2i): Boolean;
  function IsStairs(Pos: TVec2i): Boolean;
  function IsRope(Pos: TVec2i): Boolean;
  function GetCell(Cell: TVec2i): TGameObject; 
  property Player: TPlayer read FPlayer;
  property Cells[Index: TVec2i]: TGameObject read GetCell; default;
end;

TGameObject = class
protected
  //FGame: TGame;
  //FPos: TVec2D;
  //FCell: TVec2i;
  FSprite: TSprite;
  FGame: TGame;
  FPos: TVec2f;
  FFree: Boolean;
  FStairs: Boolean;
  FRope: Boolean;
  FCanBomb: Boolean;
  FTimer: Integer;
public
  constructor Create(Game: TGame; X, Y: Integer); virtual;
  destructor Destroy; override;
  procedure Draw; virtual;
  procedure Update; virtual;
  function GetCell: TVec2i; 
  function Bomb: Boolean;
  property IsFree: Boolean read FFree;
  property Stairs: Boolean read FStairs;
  property Rope: Boolean read FRope;
end;

THuman = class(TGameObject)
protected
  FOnFloor: Boolean;
  FOnStairs: Boolean;
  FAction: TAction;
public
  procedure CellMove(Dir: TVec2i; Speed: Single; IsFree: Boolean);
  procedure Right;
  procedure StartAction(Action: TAction); virtual;
  procedure StopAction(Action: TAction); virtual;
  procedure Update; override;
end;

TPlayer = class(THuman)
  constructor Create(Game: TGame; X, Y: Integer); override;
  procedure Draw; override;
  procedure StartAction(Action: TAction); override;
  procedure StopAction(Action: TAction); override;
  procedure Update; override;
end;

TEnemy = class(THuman)
  constructor Create(Game: TGame; X, Y: Integer); override;
  procedure Update; override;
end;

TNull = class(TGameObject)
  constructor Create(Game: TGame; X, Y: Integer); override;
end;

TFloor = class(TGameObject)
  constructor Create(Game: TGame; X, Y: Integer); override;
  procedure Draw; override;
  procedure Update; override;
end;

TBricks = class(TGameObject)
  constructor Create(Game: TGame; X, Y: Integer); override;
end;

TStairs = class(TGameObject)
  constructor Create(Game: TGame; X, Y: Integer); override;
end;

TRope = class(TGameObject)
  constructor Create(Game: TGame; X, Y: Integer); override;
end;

TDocument = class(TGameObject)
private
  FTaked: Boolean;
public
  constructor Create(Game: TGame; X, Y: Integer); override;
  procedure Draw; override;
  procedure Update; override;
end;

TTurnstile = class(TGameObject)
end;

TAi = class(TGameObject)
  procedure Update; override;
  procedure Draw; override;
end;

implementation

//  TGame//{{{ 
//
procedure TGame.LoadLevel(LevelName: string);//{{{
  var
    Data: TData;
    ID: Cardinal;
    I, J: Integer;
    ClassID: TGameObjectClass;
begin
  Data := TData.Create(LevelName);
  FWidth := Data.ReadInt;
  FHeight := Data.ReadInt;
  SDebug.Log(ToStr(['Loading level ', LevelName, '; ',
                    'width: ', FWidth, 
                    '; height: ', FHeight]));
  SetLength(FCells, FWidth, FHeight);
  for I := 0 to FWidth - 1 do
    for J := 0 to FHeight - 1 do begin
      ID := Data.ReadUint;
      case ID of
        1: ClassID := TFloor;
        2: ClassID := TStairs;
        3: ClassID := TRope;
        4: ClassID := TBricks;
      else
        ClassID := TNull;
      end;
      FCells[I][J] := ClassID.Create(Self, I, J);
    end;
  SetLength(FObjects, 0);
  for I := 0 to FWidth - 1 do
    for J := 0 to FHeight - 1 do begin
      ID := Data.ReadUint;
      case ID of
        1: ClassID := TAi;
        2: ClassID := TDocument;
      else
        ClassID := TGameObject;
      end;
      if ClassID <> TGameObject then begin
        SetLength(FObjects, Length(FObjects) + 1);
        FObjects[High(FObjects)] := ClassID.Create(Self, I, J);
      end;
    end;
  FPlayer := TPlayer.Create(Self, 0, 0);
  SetLength(FObjects, Length(FObjects) + 1);
  FObjects[High(FObjects)] := FPlayer;

  J := 0;
  for I := 1 to FWidth - 1 do begin
    if J = 3 then
      Break;
    if not FreeCell(Vec2i(I, 0)) then
      Continue;
    SetLength(FObjects, Length(FObjects) + 1);
    FObjects[High(FObjects)] := TEnemy.Create(Self, I, 0);
    Inc(J);
  end;

  Data.Free;
  FAi := False;
  FDocs := 4;
  SDebug.Log('Loaded');
end;
//}}}
constructor TGame.Create;//{{{
begin
  FWinApi := TWinApi.Create;
  FWindow := FWinApi.CreateWindow('Prizivnik', GWidth, GHeight);
  FWindow.Subscribe(Self);
  FSound := TSound.Create;
  FGraph := TGraph.Create(FWindow);
  SetLength(FEvents, 0);
  Log('TGame.Create');

  FTex := FGraph.LoadTex('DATA\Floor.tga');
  FSprite := CreateSprite;
  FSprite.SetTexture('DATA\Wall128.tga', 4, 4);
  FSprite.SetFramesBlock(0, 16);
  FSprite.StartAnimation(True);
  FSprite.FPS := 3;

  FLevel := '1';

  //LoadLevel('DATA\3.dlr');
  //FSprite.StartStatic;

  //FEditor := TEditor.Create(FWinApi);
end;
//}}}
destructor TGame.Destroy;//{{{
begin
  //FEditor.Free;

  FSprite.Free;

  Log('TGame.Destroy');
  FGraph.Free;
  FSound.Free;
  FWindow.Free;
  FWinApi.Free;
end;
//}}}
procedure TGame.Run;//{{{
begin
  Log('TGame.Run');
  FWinApi.Run(30);
end;
//}}}
procedure TGame.RunEditor;//{{{
  var
    Editor: TEditor;
begin
  Editor := TEditor.Create(FWinApi);
  Run;
  Editor.Free;
end;
//}}}
// //{{{
function TGame.GetTimer: Integer;
begin
  Result := TUtils.GetTimer;
end;

procedure TGame.Log(Mess: string);
begin
  SDebug.Log(Mess);
end;

procedure TGame.Assert;
begin
  SDebug.Assert(Cond, Mess);
end;

procedure TGame.PlaySound(FileName: string);
begin
  FSound.PlaySound(FileName);
end;

procedure TGame.PlayMusic(FileName: string);
begin
  FSound.PlayMusic(FileName);
end;

function TGame.GetWidth: Integer;
begin
  Result := FWindow.GetWidth;
end;

function TGame.GetHeight: Integer;
begin
  Result := FWindow.GetHeight;
end;

procedure TGame.BeginRender; inline;
begin
  FGraph.BeginRender2D(GWidth, GHeight);
end;

procedure TGame.Clear(R, G, B: Single; A: Single = 1);
begin
  FGraph.Clear(R, G, B, A);
end;

procedure TGame.EndRender;
begin
  FGraph.EndRender;
end;

procedure TGame.SetBlendMode(Mode: TBlendMode);
begin
  FGraph.SetBlendMode(Mode);
end;

function TGame.LoadTex(FileName: string): TTexture; inline;
begin
  Result := FGraph.LoadTex(FileName);
end;

function TGame.CreateSprite: TSprite; inline;
begin
  Result := FGraph.CreateSprite;
end;

procedure TGame.SetTexture(ID: TTexture); inline;
begin
  FGraph.SetTexture(ID);
end;

procedure TGame.PushMatrix; inline;
begin
  FGraph.PushMatrix;
end;

procedure TGame.PopMatrix; inline;
begin
  FGraph.PopMatrix;
end;

procedure TGame.Translate(V: TVec2f); inline;
begin
  FGraph.Translate(V);
end;
//}}}
procedure TGame.Render;//{{{
  var
    I, J: Integer;
begin
  BeginRender;
  Clear(0.0, 0.0, 0.0);
  SetBlendMode(bmAlpha);
  PushMatrix;
  Translate(Vec2f((GetWidth - (FWidth - 1)*GCell)/2, 64));
  for I := 0 to FWidth - 1 do
    for J := 0 to FHeight - 1 do begin
      FCells[I][J].Draw;
    end;
  for I := 0 to High(FObjects) do
    FObjects[I].Draw;
  PopMatrix;
  EndRender;
end;
//}}}
procedure TGame.Update;//{{{
  var
    I, J: Integer;
begin
  for I := 0 to High(FEvents) do 
    with FEvents[I] do 
      case ID of
        eiKey: if Key = Ord('1') then
                 PlaySound('DATA\Ai.wav')
               else if Key = Ord('2') then
                 PlaySound('DATA\Pila.wav')
               else if Key = VK_RIGHT then
                 FPlayer.StartAction(dirRight)
               else if Key = VK_LEFT then
                 FPlayer.StartAction(dirLeft)
               else if Key = VK_UP then
                 FPlayer.StartAction(dirUp)
               else if Key = VK_DOWN then
                 FPlayer.StartAction(dirDown)
               else if Key in [Ord('a'), Ord('A')] then
                 FPlayer.StartAction(bomLeft)
               else if Key in [Ord('s'), Ord('S')] then
                 FPlayer.StartAction(bomRight); 
        eiKeyUp: if Key = VK_RIGHT then
                   FPlayer.StopAction(dirRight)
                 else if Key = VK_LEFT then
                   FPlayer.StopAction(dirLeft)
                 else if Key = VK_UP then
                   FPlayer.StopAction(dirUp)
                 else if Key = VK_DOWN then
                   FPlayer.StopAction(dirDown);
        eiLevel: LoadLevel(FileName);
      end;
  SetLength(FEvents, 0);
  for I := 0 to High(FObjects) do begin
    FObjects[I].Update;
  end;
  for I := 0 to FWidth - 1 do
    for J := 0 to FHeight - 1 do
      FCells[I][J].Update;
end;
//}}}
procedure TGame.OnEvent;//{{{
begin
  SetLength(FEvents, Length(FEvents) + 1);
  FEvents[High(FEvents)] := Event;
end;
//}}}
procedure TGame.DoAi;//{{{
begin
  if not FAi then begin
    FAi := True;
    PlaySound('DATA\Ai2.wav');
  end;
end;//}}}
procedure TGame.Lose;//{{{
begin
  OnEvent(eString(eiLevel, 'DATA\' + FLevel + '.dlr'));
end;
//}}}
procedure TGame.Bomb(Cell: TVec2i);//{{{
begin
  if IsValidCell(Cell) then
    if Cells[Cell].Bomb then
      // FIXME: PlaySound blablabla
end;
//}}}
  procedure TGame.DecDoc;
  begin
    Dec(FDocs);
    if FDocs = 0 then begin
      FLevel := IntToStr(StrToInt(FLevel) + 1);
      OnEvent(eString(eiLevel, 'DATA\' + FLevel + '.dlr'));
    end;
  end;
function TGame.IsValidCell(Pos: TVec2i): Boolean;//{{{
begin
  Result := (Pos.X >= 0) and (Pos.Y >= 0) and
            (Pos.X < FWidth) and (Pos.Y < FHeight);
end;
//}}}
function TGame.FreeCell(Pos: TVec2i): Boolean;//{{{
begin
  if not IsValidCell(Pos) then
    Result := False
  else begin
    Result := Cells[Pos].IsFree;
   // Writeln(Cells[Pos].IsFree);
  end;
end;
//}}}
function TGame.IsStairs(Pos: TVec2i): Boolean;//{{{
begin
  if not IsValidCell(Pos) then
    Result := False
  else
    Result := Cells[Pos].Stairs;
end;
//}}}
function TGame.IsRope(Pos: TVec2i): Boolean;//{{{
begin
  if not IsValidCell(Pos) then
    Result := False
  else
    Result := Cells[Pos].Rope;
end;
//}}}
function TGame.GetCell(Cell: TVec2i): TGameObject; //{{{
begin
  Result := FCells[Cell.X][Cell.Y];
end;
//}}}
//}}}

// TGameObject//{{{
constructor TGameObject.Create;
begin
  FGame := Game;
  FPos := Vec2f(X, Y);
  FSprite := FGame.CreateSprite;
  FSprite.StartStatic;
  FFree := True;
  FStairs := False;
  FRope := False;
  FCanBomb := False;
  FTimer := 0;
end;

destructor TGameObject.Destroy;
begin
  FSprite.Free;
end;

procedure TGameObject.Draw;
begin
  FSprite.Draw(FPos*GCell, Vec2f(GCell, GCell));
end;

procedure TGameObject.Update; 
begin
end;

function TGameObject.Bomb: Boolean;
begin
  Result := False;
  if (FCanBomb) and (FTimer < FGame.GetTimer) then begin
    Result := True;
    FTimer := FGame.GetTimer + 10000;
    Writeln('Bombed!');
  end;
end;

function TGameObject.GetCell: TVec2i; 
begin
  Result := Round2f(FPos);
end;
//}}}
// THuman//{{{
procedure THuman.CellMove;
begin
  FPos += Dir * Speed;
  if not IsFree then begin
    if Dir.X = 1 then begin
      if FPos.X > GetCell.X then
        FPos.X := GetCell.X;
    end else if Dir.X = -1 then begin
      if FPos.X < GetCell.X then
        FPos.X := GetCell.X;
    end else if Dir.Y = 1 then begin
      if FPos.Y > GetCell.Y then
        FPos.Y := GetCell.Y;
    end else if Dir.Y = -1 then begin
      if FPos.Y > GetCell.Y then
        FPos.Y := GetCell.Y;
    end;
  end;
end;

procedure THuman.Right;
begin
  FPos.X += 0.2;
end;

procedure THuman.StartAction(Action: TAction);
begin
  StopAction(FAction);
  FAction := Action;
  case FAction of
    dirRight: 
      begin
        FSprite.SetFlip(False, False);
        FSprite.SetFramesBlock(0, 44);
        FSprite.StartAnimation(True);
      end;
    dirLeft: 
      begin
        FSprite.SetFlip(True, False);
        FSprite.SetFramesBlock(0, 44);
        FSprite.StartAnimation(True);
      end;
    bomLeft: FGame.Bomb(GetCell + Vec2i(-1, 1));
    bomRight: FGame.Bomb(GetCell + Vec2i(1, 1));
  end;
end;

procedure THuman.StopAction(Action: TAction);
begin
  if (FAction <> Action) or (FAction = noAction) then
    Exit;
  FSprite.StartStatic(48);
  FAction := noAction;
end;

procedure THuman.Update;
begin
  inherited Update;
  if not FGame.IsStairs(GetCell) then begin
    FOnStairs := FGame.IsStairs(GetCell + Vec2i(0, -1));
    if FGame.FreeCell(GetCell + Vec2i(0, 1)) then begin
      FOnFloor := False;
      FPos.Y += 0.098;
    end else begin
      FOnFloor := True;
      FPos.Y += 0.098;
      if FPos.Y > GetCell.Y then
        FPos.Y := GetCell.Y;
    end;
  end else begin
    FOnStairs := True;
  end;
  case FAction of
    dirRight:
      if FOnFloor or FOnStairs then begin
        CellMove(Vec2i(1, 0), 0.1, FGame.FreeCell(GetCell + Vec2i(1, 0)));
        FSprite.SetFramesBlock(0, 44);
        FSprite.StartAnimation(True);
        FSprite.SetFlip(False, False);
      end;
    dirLeft:
      if FOnFloor or FOnStairs then begin
        CellMove(Vec2i(-1, 0), 0.1, FGame.FreeCell(GetCell + Vec2i(-1, 0)));
        FSprite.SetFramesBlock(0, 44);
        FSprite.StartAnimation(True);
        FSprite.SetFlip(True, False);
      end;
    dirUp:
      if FOnStairs then begin
        //FPos.Y -= 0.1;
        CellMove(Vec2i(0, -1), 0.1, FGame.FreeCell(GetCell + Vec2i(0, -1)) and
                                    (FGame.IsStairs(GetCell + Vec2i(0, 0))));
        FPos.X := GetCell.X;
        FSprite.SetFramesBlock(0, 44);
        FSprite.StartStatic(52);
        FSprite.SetFlip(False, False);
      end;
    dirDown:
      if FOnStairs then begin
        FPos.Y += 0.1;
        FPos.X := GetCell.X;
        FSprite.SetFramesBlock(0, 44);
        FSprite.StartStatic(52);
        FSprite.SetFlip(False, False);
      end;
    else
      FSprite.StartStatic(48);
  end;
end;
//}}}
// TPlayer//{{{
constructor TPlayer.Create(Game: TGame; X, Y: Integer); 
begin
  inherited Create(Game, X, Y);
  FSprite.SetTexture('DATA\Student.tga', 8, 8);
  FSprite.SetFramesBlock(0, 44);
  FSprite.StartAnimation(True);
end;

procedure TPlayer.Draw;
begin
  FSprite.Draw(FPos*GCell + Vec2f(0, 4), Vec2f(GCell, GCell));
end;

procedure TPlayer.StartAction(Action: TAction);
begin
  inherited StartAction(Action);
end;

procedure TPlayer.StopAction(Action: TAction);
begin
  inherited StopAction(Action);
end;

procedure TPlayer.Update; 
begin
  inherited Update;
  if FAction = noAction then begin
    FSprite.StartStatic(48);
    FSprite.StartStatic(52);
  end else begin
    if (FAction = dirLeft) or (FAction = dirRight) then begin
      FSprite.SetFramesBlock(0, 44);
      FSprite.StartAnimation(True);
    end;
  end;
  if (not FGame.FreeCell(GetCell - Vec2i(1, 0))) and
     (not FGame.FreeCell(GetCell + Vec2i(1, 0))) and
     (not FGame.FreeCell(GetCell + Vec2i(0, 1))) then begin
    FGame.DoAi;
    if not FGame.FreeCell(GetCell + Vec2i(0, 0)) then
      FGame.Lose;
  end;
end;
//}}}  
constructor TEnemy.Create(Game: TGame; X, Y: Integer);
begin
  inherited Create(Game, X, Y);
  FSprite.SetTexture('DATA\Enemy.tga', 8, 8);
end;

procedure TEnemy.Update; 
begin
  case Random(100) of
    1: StartAction(dirRight);
    2: StartAction(dirLeft);
    3: StartAction(dirUp);
    4: StartAction(dirDown);
  end;
  inherited Update;
  if GetCell = FGame.Player.GetCell then
    FGame.Lose;
end;

// TNull//{{{
constructor TNull.Create(Game: TGame; X, Y: Integer);
begin
  inherited Create(Game, X, Y);
  FSprite.SetTexture('DATA\Null.tga', 1, 1);
end;
//}}}
// TFloor//{{{
constructor TFloor.Create;
begin
  inherited Create(Game, X, Y);
  FSprite.SetTexture('DATA\floor2.tga', 1, 1);
  FFree := False;
  FCanBomb := True;
end;

procedure TFloor.Draw;
begin
  if not FFree then
    inherited Draw;
end;

procedure TFloor.Update;
begin
  FFree := FTimer >= FGame.GetTimer;
  //Writeln(FTimer);
end;
//}}}
// TBricks//{{{
constructor TBricks.Create;
begin
  inherited Create(Game, X, Y);
  FSprite.SetTexture('DATA\bricks.tga', 1, 1);
  FFree := False;
end;
//}}}
// TStairs//{{{
constructor TStairs.Create;
begin
  inherited Create(Game, X, Y);
  FSprite.SetTexture('DATA\stairs.tga', 1, 1);
  FStairs := True;
end;
//}}}
// TRope//{{{
constructor TRope.Create;
begin
  inherited Create(Game, X, Y);
  FSprite.SetTexture('DATA\rope.tga', 1, 1);
  FRope := True;
end;
//}}}
// TDocument//{{{
constructor TDocument.Create(Game: TGame; X, Y: Integer);
begin
  inherited Create(Game, X, Y);
  FSprite.SetTexture('DATA\Document.tga', 1, 1);
  FTaked := False;
end;

procedure TDocument.Draw;
begin
  if not FTaked then
    inherited Draw;
end;

procedure TDocument.Update;
begin
  inherited Update;
  if (GetCell = FGame.Player.GetCell) and not FTaked then begin
    FGame.DecDoc;
    FTaked := True;
  end;
end;
//}}}
// TAi//{{{
procedure TAi.Update;
begin
  if GetCell = FGame.Player.GetCell then
    FGame.DoAi;
end;

procedure TAi.Draw;
begin
end;
//}}}

end.
