{
           32 tga!
}
unit DGraph;

interface

uses
  GL,
  GLU,
  DDebug,
  DUtils,
  DMath,
  DWinApi;

type
TSprite = class;
TGraph = class;

TTgaHeader = packed record
  idLength: Byte;
  colormapType: Byte;
  imageType: Byte;
  colormapIndex: Word;
  colormapLength: Word;
  colormapEntrySize: Byte;
  xOrigin: Word;
  yOrigin: Word;
  width: Word;
  height: Word;
  pixelSize: Byte;
  imageDesc: Byte;
end;

TProjMode = (pm2D, pm3D);
TBlendMode = (bmNo, bmAlpha, bmMul, bmAdd, bmDiffuse);

TTexture = Integer;
TTextureInfo = record
  FileName: string;
  Width, Height: Integer;
  ID: GLuint;
end;

TSprite = class
private
  FGraph: TGraph;
  FTex: TTexture;
  FFraWid: Integer;
  FFraHei: Integer;
  FFraStart: Integer;
  FFraCount: Integer;
  FFPS: Integer;
  FFrame: Integer;
  FAnimated: Boolean;
  FLooped: Boolean;
  FStart: Integer;
  FFinished: Boolean;
  FFlipX: Boolean;
  FFlipY: Boolean;
  procedure UpdateFrame;
public
  constructor Create(Graph: TGraph);
  procedure SetTexture(Tex: TTexture; FraWid, FraHei: Integer); overload;
  procedure SetTexture(Tex: string; FraWid, FraHei: Integer); overload;
  procedure SetFramesBlock(FraStart, FraCount: Integer);
  procedure SetFlip(X, Y: Boolean);
  procedure StartAnimation(Looped: Boolean);
  procedure StartStatic(Frame: Integer = 0);
  function Finished: Boolean;
  procedure Draw(Pos, Size: TVec2f; Angle: Single = 0);
  property Frame: Integer read FFrame;
  property FPS: Integer read FFPS write FFPS;
end;

TGraph = class
private
  FWindow: TWindow;
  FTextures: array of TTextureInfo;
  function LoadTexture2DToGL(W, H: Integer; P: Pointer): GLuint;
  function LoadTga(var W, H: Integer; Data: TData): GLuint;
public
  constructor Create(Window: TWindow);
  procedure BeginRender2D(Width, Height: Integer);
  procedure Clear(R, G, B: Single; A: Single = 1);
  procedure EndRender;
  procedure SetBlendMode(Mode: TBlendMode);
  function LoadTex(FileName: string): TTexture;
  function CreateSprite: TSprite; inline;
  procedure SetTexture(ID: TTexture);
  procedure SetColor(R, G, B, A: Single); inline;
  procedure Vertex2f(V: TVec2f);
  procedure TexCoord2f(V: TVec2f);
  procedure DrawRect(Pos, Size: TVec2f; Fill: Boolean = True);
  procedure PushMatrix; inline;
  procedure PopMatrix; inline;
  procedure Translate(V: TVec2f); inline;
  //procedure DrawRect();
end;

implementation

//
//  TSprite
//
procedure TSprite.UpdateFrame;
  var
    Time: Integer;
begin
  //Writeln('FAnimated: ', FAnimated);
  FFinished := False;
  if FAnimated then begin
    Time := TUtils.GetTimer;
    FFrame := (Time - FStart) * Max(1, FFPS) div 1000;
    if FLooped then
      FFrame := FFraStart + FFrame mod FFraCount
    else begin
      FFinished := FFrame >= FFraCount; 
      FFrame := Min(FFraStart + FFrame, FFraStart + FFraCount - 1);
    end;
  end;
end;

constructor TSprite.Create(Graph: TGraph);
begin
  FGraph := Graph;
  SetTexture(-1, 1, 1);
  FFPS := 30;
  FFlipX := False;
  FFlipY := False;
end;

procedure TSprite.SetTexture(Tex: TTexture; FraWid, FraHei: Integer);
begin
  FTex := Tex;
  FFraWid := Max(1, FraWid);
  FFraHei := Max(1, FraHei);
  FFraStart := 0;
  FFraCount := FraWid * FraHei;
end;

procedure TSprite.SetTexture(Tex: string; FraWid, FraHei: Integer);
begin
  SetTexture(FGraph.LoadTex(Tex), FraWid, FraHei);
end;

procedure TSprite.SetFramesBlock(FraStart, FraCount: Integer);
begin
  FFraStart := FraStart;
  FFraCount := FraCount;
end;

procedure TSprite.SetFlip(X, Y: Boolean);
begin
  FFlipX := X;
  FFlipY := Y;
end;

procedure TSprite.StartAnimation(Looped: Boolean);
begin
  FLooped := Looped;
  FAnimated := True;
  FFrame := FFraStart;
  //FStart := TUtils.GetTimer;
end;

procedure TSprite.StartStatic(Frame: Integer = 0);
begin
  FAnimated := False;
  FFrame := Frame;
end;

function TSprite.Finished: Boolean;
begin
  UpdateFrame;
  Result := FFinished;
end;

procedure TSprite.Draw(Pos, Size: TVec2f; Angle: Single = 0);
  var
    FrameSize: TVec2f;
    X, Y: Integer;
begin
  UpdateFrame;
  FGraph.SetTexture(FTex);
  Size /= 2;
  if FFlipX then
    Size.X := -Size.X;
  if FFlipY then
    Size.Y := -Size.Y; 
  FrameSize := Vec2f(1/FFraWid, 1/FFraHei);
  X := FFrame mod FFraWid;
  Y := FFrame div FFraWid;
  glBegin(GL_QUADS);
    FGraph.TexCoord2f(Vec2f(X/FFraWid, Y/FFraHei));
    FGraph.Vertex2f(Pos + Vec2f(-Size.X, -Size.Y));
    FGraph.TexCoord2f(Vec2f((X+1)/FFraWid, Y/FFraHei));
    FGraph.Vertex2f(Pos + Vec2f( Size.X, -Size.Y));
    FGraph.TexCoord2f(Vec2f((X+1)/FFraWid, (Y+1)/FFraHei));
    FGraph.Vertex2f(Pos + Vec2f( Size.X,  Size.Y));
    FGraph.TexCoord2f(Vec2f(X/FFraWid, (Y+1)/FFraHei));
    FGraph.Vertex2f(Pos + Vec2f(-Size.X,  Size.Y));
  glEnd;
end;

//
//  TGraph
//
function TGraph.LoadTexture2DToGL(W, H: Integer; P: Pointer): GLuint;
begin
  glGenTextures(1, @Result);
  glBindTexture(GL_TEXTURE_2D, Result);
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);							// set 1-byte alignment
  
  // set texture to repeat mode
  glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
  glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );

    
  gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA8, W, H, GL_RGBA, GL_UNSIGNED_BYTE, P);
  //glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, W, H, 0, GL_RGBA, GL_UNSIGNED_BYTE, P);

  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
  //glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
end;

function TGraph.LoadTga(var W, H: Integer; Data: TData): GLuint;
  var
    Header: TTgaHeader;
    I, J, Y: Integer;
    Texels: array of array[0..3] of Byte;
    Inv: Boolean;
begin
  Data.ReadVar(@Header, SizeOf(Header));
  W := Header.Width;
  H := Header.Height;
  SetLength(Texels, W*H);
  //Data.SeekAbs(SizeOf(Header));// + Header.idLength);
  Inv := Header.imageDesc and $20 = 0;
  for J := 0 to H - 1 do
    for I := 0 to W - 1 do begin
      if Inv then
        Y := H - 1 - J
      else
        Y := H;
      Texels[Y*W + I][2] := Data.ReadByte;
      Texels[Y*W + I][1] := Data.ReadByte;
      Texels[Y*W + I][0] := Data.ReadByte;
      Texels[Y*W + I][3] := Data.ReadByte;
    end;
  Result := LoadTexture2DToGL(W, H, @Texels[0]);
end;

constructor TGraph.Create(Window: TWindow);
begin
  FWindow := Window;
end;

procedure TGraph.BeginRender2D(Width, Height: Integer);
begin 
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity;
  glOrtho(0, Width, Height, 0, -1, 1);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity;
end;

procedure TGraph.Clear(R, G, B: Single; A: Single = 1);
begin
  glClearColor(R, G, B, A);
  glClear(GL_COLOR_BUFFER_BIT);
end;

procedure TGraph.EndRender;
begin
  FWindow.SwapBuffers;
end;

procedure TGraph.SetBlendMode(Mode: TBlendMode);
begin
  if Mode = bmNo then
    glDisable(GL_BLEND)
  else begin
    glEnable(GL_BLEND);
      Case Mode of
        bmAlpha:   glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        bmMul:     glBlendFunc(GL_ZERO,      GL_SRC_COLOR);
        bmAdd:     glBlendFunc(GL_ONE,       GL_ONE);
        bmDiffuse: glBlendFunc(GL_SRC_COLOR, GL_ONE_MINUS_SRC_COLOR);
      end;
  end;
end;

function TGraph.LoadTex(FileName: string): TTexture;
  var
    Data: TData;    
begin
  Result := 0;
  while (Result < Length(FTextures)) do 
    if FTextures[Result].FileName = FileName then
      Exit
    else
      Inc(Result); 

  SetLength(FTextures, Length(FTextures) + 1);
  Result := High(FTextures);
  FTextures[Result].FileName := FileName;

  Data := TData.Create(FileName);
  with FTextures[Result] do
    ID := LoadTga(Width, Height, Data);
  Data.Free;
  SDebug.Log(ToStr([
                     'TGraph.LoadTex: [file: ', FileName, '] ',
                     '[width: ', FTextures[Result].Width, '] ',
                     '[height: ', FTextures[Result].Height, '] '
                   ]));
end;

function TGraph.CreateSprite: TSprite;
begin
  Result := TSprite.Create(Self);
end;

procedure TGraph.SetTexture(ID: TTexture);
begin
  if (0 <= ID) and (ID < Length(FTextures)) then begin
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, FTextures[ID].ID);
  end else
    glDisable(GL_TEXTURE_2D); 
end;

procedure TGraph.SetColor(R, G, B, A: Single);
begin
  glColor4f(R, G, B, A);
end;

procedure TGraph.Vertex2f(V: TVec2f);
begin
  glVertex2fv(@V);
end;

procedure TGraph.TexCoord2f(V: TVec2f);
begin
  glTexCoord2fv(@V);
end;

procedure TGraph.DrawRect(Pos, Size: TVec2f; Fill: Boolean = True);
begin
  if Fill then
    glBegin(GL_QUADS)
  else
    glBegin(GL_LINE_LOOP);
    Vertex2f(Pos);
    Vertex2f(Pos + Vec2f(Size.X, 0));
    Vertex2f(Pos + Size);
    Vertex2f(Pos + Vec2f(0, Size.Y));
  glEnd;
end;

procedure TGraph.PushMatrix;
begin
  glPushMatrix;
end;

procedure TGraph.PopMatrix;
begin
  glPopMatrix;
end;

procedure TGraph.Translate(V: TVec2f);
begin
  glTranslatef(V.X, V.Y, 0.0);
end;

end.
