{$MACRO ON}
{$MODE FPC}
unit DMath;

interface

{$DEFINE FASTMATH}
{$DEFINE OPMATH}
{$DEFINE MathDir:={$IFDEF FASTMATH}inline;{$ENDIF}}

function Min(A, B: Integer): Integer;
function Max(A, B: Integer): Integer;

type
TVec2i = packed record
  case integer of
    0: (X, Y: integer);
    1: (Arr: array[0..1] of integer);
end;

TVec2f = packed record
  case integer of
    0: (X, Y: Single);
    1: (Arr: array[0..1] of Single);
end;

function Vec2i(X, Y: Integer): TVec2i; MathDir
function Round2f(V: TVec2f): TVec2i; MathDir
function Add2i(A, B: TVec2i): TVec2i; MathDir
function Sub2i(A, B: TVec2i): TVec2i; MathDir
function Mul2i(A: TVec2i; I: integer): TVec2i; MathDir
function Mul2i(A: TVec2i; S: single): TVec2f; MathDir
function Len2i(A: TVec2i): Single; MathDir

function Vec2f(X, Y: Single): TVec2f; MathDir
function Vec2f(V: TVec2i): TVec2f; MathDir
function Add2f(A, B: TVec2f): TVec2f; MathDir
function Sub2f(A, B: TVec2f): TVec2f; MathDir
function Mul2f(A: TVec2f; S: single): TVec2f; MathDir
function Len2f(A: TVec2f): Single; MathDir

{$IFDEF OPMATH}
operator = (A, B: TVec2i) Result: Boolean; inline;
operator := (A: TVec2i) Result: TVec2f; inline;
operator + (A, B: TVec2i) Result: TVec2i; inline;
operator + (A, B: TVec2f) Result: TVec2f; inline;
operator - (A, B: TVec2i) Result: TVec2i; inline;
operator - (A, B: TVec2f) Result: TVec2f; inline;
operator * (I: integer; A: TVec2i) Result: TVec2i; inline;
operator * (A: TVec2i; I: Integer) Result: TVec2i; inline;
operator * (S: Single; A: TVec2f) Result: TVec2f; inline;
operator * (A: TVec2f; S: Single) Result: TVec2f; inline;
operator div (A: TVec2i; I: Integer) Result: TVec2i; inline;
operator / (A: TVec2f; S: Single) Result: TVec2f; inline;
{$ENDIF}

implementation

function Min(A, B: Integer): Integer;
begin
  if A < B then
    Min := A
  else
    Min := B;
end;

function Max(A, B: Integer): Integer;
begin
  if A > B then
    Max := A
  else
    Max := B;
end;

// TVec2i//{{{
function Vec2i(X, Y: Integer): TVec2i; MathDir
begin
  Vec2i.X := X;
  Vec2i.Y := Y;
end;

function Round2f(V: TVec2f): TVec2i; MathDir
begin
  Round2f.X := Round(V.X);
  Round2f.Y := Round(V.Y);
end;

function Add2i(A, B: TVec2i): TVec2i; MathDir
begin
  Add2i.X := A.X + B.X;
  Add2i.Y := A.Y + B.Y;
end;

function Sub2i(A, B: TVec2i): TVec2i; MathDir
begin
  Sub2i.X := A.X - B.X;
  Sub2i.Y := A.Y - B.Y;
end;

function Mul2i(A: TVec2i; I: integer): TVec2i; MathDir
begin
  Mul2i.X := I * A.X;
  Mul2i.Y := I * A.Y;
end;

function Mul2i(A: TVec2i; S: single): TVec2f; MathDir
begin
  Mul2i.X := S * A.X;
  Mul2i.Y := S * A.Y;
end;

function Len2i(A: TVec2i): Single; MathDir
begin
  Len2i := Sqrt(Sqr(A.X) + Sqr(A.Y));
end;
//}}}
// TVec2f//{{{
function Vec2f(X, Y: Single): TVec2f; MathDir
begin
  Vec2f.X := X;
  Vec2f.Y := Y;
end;

function Vec2f(V: TVec2i): TVec2f; MathDir
begin
  Vec2f := Vec2f(V.X, V.Y);
end;

function Add2f(A, B: TVec2f): TVec2f; MathDir
begin
  Add2f.X := A.X + B.X;
  Add2f.Y := A.Y + B.Y;
end;

function Sub2f(A, B: TVec2f): TVec2f; MathDir
begin
  Sub2f.X := A.X - B.X;
  Sub2f.Y := A.Y - B.Y;
end;

function Mul2f(A: TVec2f; S: single): TVec2f; MathDir
begin
  Mul2f.X := S * A.X;
  Mul2f.Y := S * A.Y;
end;

function Len2f(A: TVec2f): Single; MathDir
begin
  Len2f := Sqrt(Sqr(A.X) + Sqr(A.Y));
end;//}}}

// operators//{{{
{$IFDEF OPMATH}
operator = (A, B: TVec2i) Result: Boolean; inline;
begin
  Result := (A.X = B.X) and (A.Y = B.Y);
end;

operator := (A: TVec2i) Result: TVec2f; inline;
begin
  Result := Vec2f(A.X, A.Y)
end;

operator + (A, B: TVec2i) Result: TVec2i; inline;
begin
  Result := Add2i(A, B);
end;

operator + (A, B: TVec2f) Result: TVec2f; inline;
begin
  Result := Add2f(A, B);
end;

operator - (A, B: TVec2i) Result: TVec2i; inline;
begin
  Result := Sub2i(A, B);
end;

operator - (A, B: TVec2f) Result: TVec2f; inline;
begin
  Result := Sub2f(A, B);
end;

operator * (I: integer; A: TVec2i) Result: TVec2i; inline;
begin
  Result := Mul2i(A, I)
end;

operator * (A: TVec2i; I: Integer) Result: TVec2i; inline;
begin
  Result := Mul2i(A, I)
end;

operator * (S: Single; A: TVec2f) Result: TVec2f; inline;
begin
  Result := Mul2f(A, S)
end;

operator * (A: TVec2f; S: Single) Result: TVec2f; inline;
begin
  Result := Mul2f(A, S)
end;

operator div (A: TVec2i; I: Integer) Result: TVec2i; inline;
begin
  Result := Vec2i(A.X div I, A.Y div I);
end;

operator / (A: TVec2f; S: Single) Result: TVec2f; inline;
begin
  Result := Mul2f(A, 1/S);
end;
{$ENDIF OPMATH}
//}}}
end.
