{$MACRO on}
unit DUtils;

interface

uses
  windows;

Type
TUtils = class
  class function GetTimer: Integer;
end;

generic TList<TItem> = class(TObject)
var private
  FItems: Array Of TItem;
public
  // SetLength
  procedure S(Len: Integer); inline;
  // High
  function H: Integer; inline;
  // Length  
  function L: Integer; inline;
  procedure Add(Item: TItem);
  procedure Del(Item: TItem; All: Boolean = True);
  function Find(Item: TItem): Integer;
  function GetItem(Index: Integer): TItem; //inline;
  procedure SetItem(Index: Integer; Item: TItem); //inline;
  property Items[Index: Integer]: TItem read GetItem write SetItem;
end;
//TListOfInteger = specialize _TList<Integer>;

TData = class
private
  FBytes: array of Byte;
  FPos: Integer;
public
  constructor Create; overload;
  constructor Create(FileName: string); overload;
  procedure Clear;
  procedure ReadFromFile(FileName: string);
  procedure WriteToFile(FileName: string);
  procedure WriteVar(P: Pointer; Size: Integer);
  procedure WriteInt(I: Integer);
  procedure Rewind;
  procedure SeekAbs(NewPos: Integer);
  procedure ReadVar(P: Pointer; Size: Integer);
  function ReadInt: Integer;
  function ReadUint: Cardinal;
  function ReadByte: Byte;
  function GetPtr: Pointer;
end;

Function IntToStr(Value: integer): string;
Function StrToInt(Value: string): integer;
Function StrToIntDef(Value: String; Def: integer): integer;
Function StrToFloat(S: String): Single;
Function FloatToStr(F: Single): String;
function CharArrToStr(const C : Array of Char) : String;
Function HexToInt(S: String): Integer;
Function IntToHex(Value: Integer; Digits: Integer = 0): String;
Function FileExist(Filename: string): boolean;
Procedure CreateFile(FileName: string);
Procedure DeleteFile(Filename: string);
Function GetTimer: integer;
//     
Function AnsiLowerCase(const S: string): string;

function ToStr(V: Array Of Const): String;

implementation

class function TUtils.GetTimer: Integer;
var
  T, F: LARGE_INTEGER;
begin
  QueryPerformanceFrequency(int64(F));
  QueryPerformanceCounter(int64(T));
  Result:=Trunc(1000*T.QuadPart/F.QuadPart);
end;


//  generic TList//{{{
//
procedure TList.S(Len: Integer);
begin
  {System.}SetLength(FItems, Len);
end;

function TList.H: Integer;
begin
  Result := {System.}High(FItems);
end;

function TList.L: Integer;
begin
  Result := {System.}Length(FItems);
end;

{$DEFINE Length:=L}
{$DEFINE High:=H}
{$DEFINE SetLength:=S}
procedure TList.Add(Item: TItem);
begin
  SetLength(Length + 1);
  FItems[High] := Item;
end;

procedure TList.Del(Item: TItem; All: Boolean = True);
  var
    I: Integer;
begin
  I := 0;
  while I < High do
    if FItems[I] = Item then begin
      FItems[I] := FItems[High];
      SetLength(Length - 1);
      if not All then
        Exit;
    end else
      Inc(I);
end;

function TList.Find(Item: TItem): Integer;
  var
    I: Integer;
begin
  for I := 0 to High do
    if FItems[I] = Item then begin
      Result := I;
      Exit;
    end;
  Result := -1;
end;

function TList.GetItem(Index: Integer): TItem;
begin
  Result := FItems[Index];
end;

procedure TList.SetItem(Index: Integer; Item: TItem);
begin
  FItems[Index] := Item;
end;

{$UNDEF Length}
{$UNDEF High}
{$UNDEF SetLength}
//}}}
//{{{ TData
constructor TData.Create;
begin
  Clear;
  Rewind;
end;

constructor TData.Create(FileName: string);
begin
  Create;
  ReadFromFile(FileName);
end;

procedure TData.Clear;
begin
  SetLength(FBytes, 0);
end;

procedure TData.ReadFromFile(FileName: string);
  var
    F: file of Byte;
begin
  Assign(F, FileName);
  {$I-}
  Reset(F);
  {$I+}
  if IOResult <> 0 then
    Exit;
  SetLength(FBytes, Length(FBytes) + FileSize(F));
  BlockRead(F, FBytes[Length(FBytes) - FileSize(F)], FileSize(F));
  Close(F);
end;

procedure TData.WriteToFile(FileName: string);
  var
    F: file of Byte;
begin
  Assign(F, FileName);
  {$I-}
  Rewrite(F);
  {$I+}
  if IOResult <> 0 then
    Exit;
  SetLength(FBytes, Length(FBytes) + FileSize(F));
  BlockWrite(F, FBytes[0], Length(FBytes));
  Close(F);
end;

procedure TData.WriteVar(P: Pointer; Size: Integer);
begin
  SetLength(FBytes, Length(FBytes) + Size);
  Move(P^, FBytes[Length(FBytes) - Size], Size);
end;

procedure TData.WriteInt(I: Integer);
begin
  WriteVar(@I, SizeOf(Integer));
end;

procedure TData.Rewind;
begin
  FPos := 0;
end;

procedure TData.SeekAbs(NewPos: Integer);
begin
  FPos := NewPos;
end;

procedure TData.ReadVar(P: Pointer; Size: Integer);
begin
  if FPos + Size <= Length(FBytes) then begin
    Move(FBytes[FPos], P^, Size);
    FPos += Size;
  end;
end;

function TData.ReadInt: Integer;
begin
  ReadVar(@Result, SizeOf(Integer));
end;

function TData.ReadUint: Cardinal;
begin
  ReadVar(@Result, SizeOf(Cardinal));
end;

function TData.ReadByte: Byte;
begin
  ReadVar(@Result, SizeOf(Byte))
end;

function TData.GetPtr: Pointer;
begin
  Result := @FBytes[FPos];
end;
//}}}
Function InttoStr;
begin
  Str(Value, Result);
end;

Function StrtoInt;
  Var
    Code: integer;
begin
  Val(Value, Result, Code);
end;

Function StrToIntDef(Value: String; Def: integer): integer;
  Var
    Code: integer;
begin
  Val(Value, Result, Code);
    If Code <> 0 then
      Result := Def;
end;

Function StrToFloat(S: String): Single;
  Var
    Code: Integer;
begin
    If Pos('.', S) <> 0 then begin
      //Insert(',', S, Pos('.', S));
      //Delete(S, Pos('.', S), 1);
    end;
  Val(S, Result, Code);
end;

Function FloatToStr(F: Single): String;
begin
  Str(F, Result);
    If Pos(',', Result) <> 0 then begin
      Insert('.', Result, Pos(',', Result));
      Delete(Result, Pos(',', Result), 1);
    end;
end;

Function HexDigitToInt(C: Char): Byte;
begin
  if C in ['A'..'F'] then
    Result := Ord(C) - Ord('A') + 10
  else if C in ['a'..'f'] then
    Result := Ord(C) - Ord('a') + 10
  else if C in ['0'..'9'] then
    Result := Ord(C) - Ord('0')
  else
    Result := 0;
end;

Function HexToInt(S: String): Integer;
begin
  Result := 0;
  while Length(S) > 0 do begin
    Result := Result*16 + HexDigitToInt(S[1]);
    Delete(S, 1, 1);
  end;
end;

const
   HexDigits: array[0..15] of char = '0123456789ABCDEF';

function IntToHex;
  var 
    i: integer;
begin
 SetLength(result, digits);
 for i := 0 to digits - 1 do
  begin
   result[digits - i] := HexDigits[value and 15];
   value := value shr 4;
  end ;
 while value <> 0 do begin
   result := HexDigits[value and 15] + result;
   value := value shr 4;
 end;
end ;

function CharArrToStr(const C : Array of Char) : String;
var I : Integer;
begin
  // convert the array of characters to a String
  I :=0;
  result :='';
  while C[i] <> #0 do
  begin
    result := result + C[I];
    Inc(I);
  end;
end;

Function FileExist;
  Var
    F: File;
begin
  FileMode := 64;
  AssignFile(F, Filename);
  {$I-}
  Reset(F);
  {$I+}
  Result := IOResult = 0;
    If Result then
      CloseFile(F);
end;

Procedure CreateFile;
  Var
    F: TextFile;
begin
    If FileExist(FileName) then
      Exit;
  Assign(F, FileName);
  {$I-}
  Rewrite(F);
  {$I+}
    If IOResult = 0 then
      Close(F);
end;

Procedure DeleteFile;
begin
    If not FileExist(Filename) then
      Exit;
  Windows.DeleteFile(PAnsiChar(Filename));
end;

Function GetTimer;
var
  T, F: LARGE_INTEGER;
begin
  QueryPerformanceFrequency(int64(F));
  QueryPerformanceCounter(int64(T));
  Result:=Trunc(1000*T.QuadPart/F.QuadPart);
end;

Function AnsiLowerCase(Const S: String): String;
  Var
    Len: Integer;
begin
  Len := Length(S);
  SetString(Result, PChar(S), Len);
    If Len > 0 then
      CharLowerBuff(Pointer(Result), Len);
end;

Function ToStr(V: Array Of Const): String;
  Var
    I: Integer;
begin
  Result := '';
  for I := 0 to High(V) do
    case V[I].vType of 
      vtInteger: Result += IntToStr(V[I].vInteger);
      vtBoolean: Result += IntToStr(Ord(V[I].vBoolean));
      vtChar: Result += V[I].vChar;
      vtString: Result += V[I].vString^;
      vtPointer: Result += '@' + IntToStr(LongInt(V[I].vPointer));
      vtPChar: Result += V[I].vPChar;
      vtObject:
        begin
          Result += '<' + V[I].vObject.ClassName + '>';
        end;
      vtClass: Result += '<' + V[I].vClass.ClassName + '>';
      vtAnsiString: Result += AnsiString(V[I].vAnsiString);
    else
      Result += '?';
    end;
end;

end.
