unit DWinApi;

interface

uses
  windows,
  //messages,
  GL,
  DUtils,
  DDebug,
  DMath,
  DEvents;

Type
TWinApi = class;

TWinApp = class
  procedure Render; virtual;
  procedure Update; virtual;
  procedure OnEvent(const Event: TEvent); virtual;
end;

TWindow = class
private
  FWinApi: TWinApi;
  FApp: TWinApp;
  //      ,    Win32
  FHandle: HWND;
  FDC: HDC;
  FRC: HGLRC;
  function WndProc(Msg: Cardinal; wParam, lParam: Integer): Boolean;
  property Handle: HWND read FHandle write FHandle;
  property DC: HDC read FDC write FDC;
  property RC: HGLRC read FRC write FRC;
  class function CreateWindowClass(Name: String): Boolean;
  function CreatePixelFormat(ColorBits, DepthBits, StencilBits: Integer): Boolean;
public
  constructor Create(WinApi: TWinApi; Caption: String; X, Y, Width, Height: Integer);
  destructor Destroy; override;
  procedure Subscribe(App: TWinApp);
  procedure Render;
  procedure Update;
  procedure SwapBuffers; inline;
  function GetWidth: Integer; inline;
  function GetHeight: Integer; inline;
  function GetSize: TVec2i;
  function GetPos: TVec2i;
  function GetMouse: TVec2i; inline;
  property WinApi: TWinApi read FWinApi;
  property Pos: TVec2i read GetPos;
  property Size: TVec2i read GetSize;
  property Mouse: TVec2i read GetMouse;
end;

TWinApi = class
private
  FWindows: array of TWindow;
  FExiting: Boolean;
public
  constructor Create;
  destructor Destroy; override;
  function CreateWindow(Caption: String; W, H: Integer): TWindow;
  procedure Run(UpdInterval: Integer);
  procedure ShutDown;
  function SWidth: Integer;
  function SHeight: Integer; 
  function GetMouse: TVec2i;
  property Mouse: TVec2i read GetMouse;
end;

//     ext,    
Function GetSaveFile(Window: TWindow; ext: string = '*'): string;
//    ext,    
Function GetLoadFile(Window: TWindow; ext: string = '*'): string;

implementation

//Var
//  Windows: Array Of TWindow;

//  ,  //{{{
function WndProc(hWnd: HWND; Msg: Cardinal; wParam: Integer; lParam: Integer): Integer; stdcall;
begin
  Result := 0;
  {
  // Windows      .
  //       ,   .
  //   hWnd,   .
  //        Windows, : 
  for I := 0 to High(Windows) do
    if Windows[I].Handle = hWnd then // ,  
      if not Windows[I].WndProc(Msg, wParam, lParam) then // ?
        Break
      else
        Exit;
  }
  if TWindow(GetWindowLong(hWnd, GWL_USERDATA)) <> nil then
    if TWindow(GetWindowLong(hWnd, GWL_USERDATA)).
               WndProc(Msg, wParam, lParam) then
      Exit;
  Result := DefWindowProc(hWnd, Msg, wParam, lParam);
end;
//}}}
//  TWinApp//{{{
//
procedure TWinApp.Render;
begin
end;

procedure TWinApp.Update;
begin
end;

procedure TWinApp.OnEvent;
begin
end;
//}}}
//  TWindow//{{{
//

// ,      Name//{{{
class function TWindow.CreateWindowClass(Name: String): Boolean;
  Var
    Wnd: TWNDCLASSEXA;
begin
  Result := False;
  FillChar(wnd, SizeOf(wnd), 0);
    With wnd do begin
      cbSize        := SizeOf(wnd);
      style         := CS_HREDRAW or
                       CS_VREDRAW or
                       CS_OWNDC;
      lpfnWndProc   := @DWinApi.WndProc;
      hCursor       := LoadCursor(0, IDC_ARROW);
      lpszClassName := PAnsiChar(Name);
    end;
    If RegisterClassEx(wnd) = 0 then
      Exit;
  Result := True;
  //Log.PrintLn('CreateWindowClass("' + Name + '");');
end;
//}}}
// ,      //{{{
function TWindow.CreatePixelFormat(ColorBits, DepthBits, StencilBits: Integer): Boolean;
  Var
    Pfd: PIXELFORMATDESCRIPTOR;
    PixelFormat: Integer;
begin
  Result := False;
  FillChar(pfd, SizeOf(pfd), 0);
    With pfd do begin
      nSize        := SizeOf(TPIXELFORMATDESCRIPTOR);
      nVersion     := 2;
      dwFlags      := PFD_DRAW_TO_WINDOW or
                      PFD_SUPPORT_OPENGL or
                      PFD_DOUBLEBUFFER;
      iPixelType   := PFD_TYPE_RGBA;
      cColorBits   := ColorBits;
      cDepthBits   := DepthBits;
      cStencilBits := StencilBits;
      iLayerType   := PFD_MAIN_PLANE;
    end;
  PixelFormat := ChoosePixelFormat(DC, @pfd);
    If (PixelFormat = 0) Or Not SetPixelFormat(DC, PixelFormat, @Pfd) then
      Exit;
  Result := True;
  //Log.PrintLn('CreatePixelFormat(Color = ' + IntToStr(ColorBits) + ', ' +
  //                              'Depth = ' + IntToStr(DepthBits) + ', ' +
  //                              'Stencil = ' + IntToStr(StencilBits) + ');');
end;
//}}}
constructor TWindow.Create;//{{{
  Var
    dmScreenSettings : DEVMODE;
    dwStyle : DWORD;
    //dwExStyle : DWORD;
    Rect  : TRect;
begin
  //InitOpenGL;
  FWinApi := WinApi;

  //if Not CreateWindowClass('OpenGL') then begin
  //  SDebug.Log('Error[TWindow.Create]: CreateWindowClass');
  //  Exit;
  //end;
  
  FHandle := CreateWindowEx(WS_EX_APPWINDOW or WS_EX_WINDOWEDGE, 
                            'OpenGL', 
                            PChar(Caption),
                            WS_POPUP or WS_CAPTION, 
                            X, Y, Width, Height, 0, 0, 0, nil);
  if FHandle = 0 then begin
    SDebug.Log('Error[TWindow.Create]: CreateWindowEx');
    Exit;
  end;
  
  DC := GetDC(FHandle);
  if (DC = 0) Or Not CreatePixelFormat(32, 24, 8) then begin
    SDebug.Log('Error[TWindow.Create]: CreatePixelFormat');
    Exit;
  end;
  
  RC := wglCreateContext(DC);
  if (RC = 0) Or Not wglMakeCurrent(DC, RC) then begin
    SDebug.Log('Error[TWindow.Create]: wglCreateContext or wglMakeCurrent');
    Exit;
  end;

  SetWindowLong(FHandle, 0, WS_VISIBLE);
  Rect.Left   := 0;
  Rect.Top    := 0;
  Rect.Right  := width;
  Rect.Bottom := height;
  AdjustWindowRect(Rect, dwStyle, False);
  with Rect do
    MoveWindow(FHandle, 0, 0, Right - Left, Bottom - Top, True);

  ShowCursor(True);
  SetForegroundWindow(FHandle);
  ShowWindow(FHandle, SW_SHOW);
  SetFocus(FHandle);
  UpdateWindow(FHandle);
  //SetWindowLong(FHandle, GWL_STYLE, (WS_CAPTION*Ord(Not FullScreen)) or WS_VISIBLE);
  SetWindowLong(FHandle, GWL_STYLE,  WS_CAPTION or WS_VISIBLE);  
  SetWindowPos(FHandle, 0, 400, 200, Width, Height, SWP_FRAMECHANGED or SWP_NOOWNERZORDER);
  glViewport(0, 0, Width, Height);


  SetWindowLong(FHandle, GWL_USERDATA, Cardinal(Self)); 
  //SetLength(Windows, Length(Windows) + 1);
  //Windows[High(Windows)] := Self;
  SDebug.Log('TWindow.Create');
end;
//}}}
destructor TWindow.Destroy;//{{{
  var
    I: Integer;
begin
  {FIXME: }
  // FWinApi.DelWindow(...)
end;
//}}}
procedure TWindow.Subscribe;//{{{
begin
  FApp := App;
end;
//}}}
function TWindow.WndProc(Msg: Cardinal; wParam, lParam: Integer): Boolean;//{{{
  var
    Event: TEvent;
begin
  //SDebug.Log('TWindow.WndProc in');
  if (Msg = WM_KEYDOWN) and (wParam = 13) then
    FWinApi.ShutDown;
    
  Event.ID := eiNo;
  if Msg = WM_KEYDOWN then begin
    Event.ID := eiKey;
    Event.Key := wParam;
  end else if Msg = WM_KEYUP then begin
    Event.ID := eiKeyUp;
    Event.Key := wParam;
  end else if (Msg = WM_LBUTTONDOWN) or
         (Msg = WM_RBUTTONDOWN) or
         (Msg = WM_MBUTTONDOWN) or
         (Msg = WM_LBUTTONDBLCLK) then begin
    Event.ID := eiClick;
    Event.Mouse := Mouse;
    if Msg = WM_LBUTTONDOWN then
      Event.Button := miLeft
    else if Msg = WM_MBUTTONDOWN then
      Event.Button := miMiddle
    else if Msg = WM_RBUTTONDOWN then
      Event.Button := miRight
    else if Msg = WM_LBUTTONDBLCLK then
      Event.Button := miDLeft;
  end;
  if Event.ID <> eiNo then
    FApp.OnEvent(Event);
  Result := False;
  //SDebug.Log('TWindow.WndProc out');
end;
//}}}
procedure TWindow.Render;//{{{
begin
  if FApp <> nil then
    FApp.Render;
end;
//}}}
procedure TWindow.Update;//{{{
begin
  if FApp <> nil then
    FApp.Update;
end;
//}}}
procedure TWindow.SwapBuffers;//{{{
begin
  Windows.SwapBuffers(FDC);
end;
//}}}
function TWindow.GetWidth: Integer;//{{{
begin
  Result := GetSize.X;
end;
//}}}
function TWindow.GetHeight: Integer;//{{{
begin
  Result := GetSize.Y;
end;
//}}}
function TWindow.GetSize: TVec2i;//{{{
  var
    Rect: TRect;
begin
  GetWindowRect(FHandle, Rect);
  Result := Vec2i(Rect.Right - Rect.Left, Rect.Bottom - Rect.Top);
end;
//}}}
function TWindow.GetPos: TVec2i;//{{{
var
  Rect: TRect;
begin
  GetWindowRect(FHandle, Rect);
  Result := Vec2i(Rect.Left, Rect.Top);
end;
//}}}
function TWindow.GetMouse: TVec2i; inline;//{{{
begin
  Result := FWinApi.Mouse - GetPos;
end;
//}}}//}}}
//  TWinApi//{{{
//
constructor TWinApi.Create;
begin
  SetLength(FWindows, 0);

  if not TWindow.CreateWindowClass('OpenGL') then begin
    SDebug.Log('Error[TWindow.Create]: CreateWindowClass');
    Exit;
  end;
end;

destructor TWinApi.Destroy;
begin
  //Windows.Free;
end;

function TWinApi.CreateWindow(Caption: String; W, H: Integer): TWindow;
begin
  Result := TWindow.Create(Self,
                           Caption, 
                           (SWidth - W) div 2,
                           (SHeight - H) div 2,
                           W,
                           H);
  SetLength(FWindows, Length(FWindows) + 1);
  FWindows[High(FWindows)] := Result;
end;

procedure TWinApi.Run;//{{{
  var
    Flag: Boolean;
    //Exiting: Boolean;
    msg: TMsg;
    Time, LastUpdate: Integer;
    I: Integer;
begin
  FExiting := False;
  LastUpdate := TUtils.GetTimer;
  while not FExiting do begin
    Flag := False;
    Time := TUtils.GetTimer;
    while Time - LastUpdate > UpdInterval do begin
      for I := 0 to High(FWindows) do
        FWindows[I].Update;
      flag := True;
      LastUpdate += UpdInterval;
    end;
    if Flag then
      for I := 0 to High(FWindows) do
        FWindows[I].Render;
    while PeekMessage(msg, 0, 0, 0, PM_REMOVE) do begin
      TranslateMessage(msg);
      DispatchMessage(msg);
    end;
  end;
end;
//}}}
procedure TWinApi.ShutDown;
begin
  FExiting := True;
end;

function TWinApi.SWidth: Integer;
begin
  Result := 1024;
end;

function TWinApi.SHeight: Integer; 
begin
  Result := 768;
end;

function TWinApi.GetMouse;
  Var
    Point: TPoint;
begin
  GetCursorPos(Point);
  Result := Vec2i(Point.X, Point.Y);
end;
//}}}

// Dlg//{{{
Const
  OFN_PATHMUSTEXIST = $00000800;
  OFN_HIDEREADONLY = $00000004;
  OFN_FILEMUSTEXIST = $00001000;

  commdlg32 = 'comdlg32.dll';

Type
  tagOFNA = packed record
    lStructSize: DWORD;
    hWndOwner: HWND;
    hInstance: HINST;
    lpstrFilter: PAnsiChar;
    lpstrCustomFilter: PAnsiChar;
    nMaxCustFilter: DWORD;
    nFilterIndex: DWORD;
    lpstrFile: PAnsiChar;
    nMaxFile: DWORD;
    lpstrFileTitle: PAnsiChar;
    nMaxFileTitle: DWORD;
    lpstrInitialDir: PAnsiChar;
    lpstrTitle: PAnsiChar;
    Flags: DWORD;
    nFileOffset: Word;
    nFileExtension: Word;
    lpstrDefExt: PAnsiChar;
    lCustData: LPARAM;
    lpfnHook: function(Wnd: HWND; Msg: UINT; wParam: WPARAM; lParam: LPARAM): UINT stdcall;
    lpTemplateName: PAnsiChar;
    pvReserved: Pointer;
    dwReserved: DWORD;
    FlagsEx: DWORD;
  end;
  tagOFNW = packed record
    lStructSize: DWORD;
    hWndOwner: HWND;
    hInstance: HINST;
    lpstrFilter: PWideChar;
    lpstrCustomFilter: PWideChar;
    nMaxCustFilter: DWORD;
    nFilterIndex: DWORD;
    lpstrFile: PWideChar;
    nMaxFile: DWORD;
    lpstrFileTitle: PWideChar;
    nMaxFileTitle: DWORD;
    lpstrInitialDir: PWideChar;
    lpstrTitle: PWideChar;
    Flags: DWORD;
    nFileOffset: Word;
    nFileExtension: Word;
    lpstrDefExt: PWideChar;
    lCustData: LPARAM;
    lpfnHook: function(Wnd: HWND; Msg: UINT; wParam: WPARAM; lParam: LPARAM): UINT stdcall;
    lpTemplateName: PWideChar;
    pvReserved: Pointer;
    dwReserved: DWORD;
    FlagsEx: DWORD;
  end;
  tagOFN = tagOFNA;
  TOpenFilenameA = tagOFNA;
  TOpenFilenameW = tagOFNW;
  TOpenFilename = TOpenFilenameA;

function GetSaveFileName(var OpenFile: TOpenFilename): Bool; stdcall; external commdlg32 name 'GetSaveFileNameA';
function GetOpenFileName(var OpenFile: TOpenFilename): Bool; stdcall; external commdlg32 name 'GetOpenFileNameA';

Function GetSaveFile;
  Var
  SaveDialog: TOpenFilename;
begin
  ZeroMemory(@SaveDialog, SizeOf(SaveDialog));
    with SaveDialog do begin
    //  
      lStructSize   := SizeOf(SaveDialog);
    //   (  )
      hWndOwner     := Window.Handle;
    //   ( )
      lpstrFilter   := PAnsiChar(' ' + '*.' + ext + #0 + '*.' + ext + #0#0);
    //       ,   ...
      lpstrDefExt   := PAnsiChar(ext);
    //      
    //    "  ", ..      ;)
      Flags         := OFN_PATHMUSTEXIST or OFN_HIDEREADONLY;
    //      ()   
      nMaxFile      := 250;
      nMaxFileTitle := nMaxFile;
      GetMem(lpstrFile, nMaxFile);
      ZeroMemory(lpstrFile, nMaxFile);
    end;
    If GetSaveFileName(SaveDialog) then
      Result := SaveDialog.lpstrFile
    else
      Result := '';
  //FreeMem(SaveDialog.lpstrFile);
end;

Function GetLoadFile;
  Var
  LoadDialog: TOpenFilename;
begin
  ZeroMemory(@LoadDialog, SizeOf(LoadDialog));
    with LoadDialog do begin
    //  
      lStructSize   := SizeOf(LoadDialog);
    //   (  )
      hWndOwner     := Window.Handle;
    //   ( )
      lpstrFilter   := PAnsiChar(' ' + '*.' + ext + #0 + '*.' + ext + #0#0);
    //       ,   ...
      lpstrDefExt   := PAnsiChar(ext);
    //      
    //    "  ", ..      ;)
      Flags         := OFN_PATHMUSTEXIST or OFN_HIDEREADONLY or OFN_FILEMUSTEXIST;
    //      ()   
      nMaxFile      := 250;
      nMaxFileTitle := nMaxFile;
      GetMem(lpstrFile, nMaxFile);
      ZeroMemory(lpstrFile, nMaxFile);
    end;
    If GetOpenFileName(LoadDialog) then
      Result := LoadDialog.lpstrFile
    else
      Result := '';
end;
//}}}

end.
